!! Search PDB for a fragment with low RMSD to a target fragment,
!! with maaximum number of H-bonds and zero collisions to an atoms
!! set, and zero H-bonding groups blocked.
!! CB 26-May-2009
!!-----------------------------------------------------------------------
program  anchor
  use names
  use bdc
  use vectormath, only: superimp, moveall
  implicit none
  integer,parameter :: STACKDEPTH=5, munit=49, lunit=15
  real,parameter :: rmsdcut=6.0, rmsdmin=0.1, MAXD=13., wiggleroom=0.4
  character(len=1000) :: listname, pdbfile, aline, lassofile,ligandfile,bdcfile
  character(len=1000) :: moviefile="anchor_movie.pdb"
  integer :: jarg,iargc,ios, ires,nres,iat,nat,jat
  integer :: i,j,k,ht,slen,srec,imdl,irec,nseg,iu=0,nrmsd,ii,jj
  real,dimension(:,:),pointer :: lasso,segwindow,cawindow,calasso
  real,dimension(:,:),pointer :: segment,segcenter
  real,dimension(:,:),allocatable :: ligand_coords
  integer,dimension(:),allocatable :: ligand_htype
  real,dimension(3) :: vec, atvec, vx, wigglevec, bestwiggle
  real,dimension(3,3) :: mtx
  !! lasso stack
  type stacktype
    real,dimension(:,:),pointer :: coords
    real :: score
    integer :: irec
  end type stacktype
  type (stacktype), dimension(STACKDEPTH+1) :: lassostack
  real :: hbscore, hbest, rmsd,lastbest,dd,dist, bestscore
  !!---------------- command line
  jarg = iargc()
  if (jarg < 3) then
    write(*,*) "Usage: xanchor lassocoords receptorcoords binaryPDB"
    write(*,*) "lassocoords = preliminary coordinates for a bound peptide"
    write(*,*) "receptorcoords = coords to be avoided."
    write(*,*) "binaryPDB = database backbone coordinates from pdb2bin.f90"
    write(*,*) "Output VMD movie to ",trim(moviefile)
    stop 'anchor.f90 v.  Tue May 26 15:22:36 PDT 2009'
  endif
  call getarg(1,lassofile)
  call getarg(2,ligandfile)
  call getarg(3,bdcfile)
  call bdc_init(bdcfile)
  !! ----- open input lasso file
  open(11,file=lassofile,status='old',form='formatted',iostat=ios)
  if (ios/=0) stop 'anchor.f90:: ERROR lassofile not found'
  !! ----- open file for VMD movie
  open(munit,file=trim(moviefile),status='replace',form='formatted',iostat=ios)
  if (ios/=0) stop 'anchor:: error opening output file. Permissions?'
  !! ----------------- read preliminary lasso 
  ires = 0
  do 
    read(11,'(a)',iostat=ios) aline
    if (ios/=0) exit
    if (aline(1:5)/="ATOM ") cycle
    select case (aline(13:16))
    case (" N  ")
      ires = ires + 1
    end select
  enddo
  nres = ires
  write(*,'(i9," Residues read from ",a)') nres, trim(lassofile)
  allocate(lasso(3,5*nres),stat=ios); if (ios/=0) stop 'Error allocaintg lasso'
  allocate(segwindow(3,5*nres),stat=ios); if (ios/=0) stop 'Error allocating segwindow'
  allocate(segcenter(3,5*nres),stat=ios); if (ios/=0) stop 'Error allocating segcenter'
  allocate(calasso(3,nres),stat=ios); if (ios/=0) stop 'Error allocating calasso'
  allocate(cawindow(3,nres),stat=ios); if (ios/=0) stop 'Error allocating cawindow'
  rewind(11)
  lasso = 999.
  ires = 0
  do 
    read(11,'(a)',iostat=ios) aline
    if (ios/=0) exit
    if (aline(1:5)/="ATOM ") cycle
    !ATOM      1  N   ILE    62      28.865  53.961  42.655  1.00  0.00           N1+
    select case (aline(13:16))
    case (" N  ")
      ires = ires + 1
      read(aline(31:54),'(3f8.3)',iostat=ios) lasso(1:3,(ires-1)*5+1)
      if (ios/=0) stop "BUG parsing ATOM line"
    case (" CA ")
      read(aline(31:54),'(3f8.3)',iostat=ios) lasso(1:3,(ires-1)*5+2)
      if (ios/=0) stop "BUG parsing ATOM line"
    case (" C  ")
      read(aline(31:54),'(3f8.3)',iostat=ios) lasso(1:3,(ires-1)*5+3)
      if (ios/=0) stop "BUG parsing ATOM line"
    case (" O  ")
      read(aline(31:54),'(3f8.3)',iostat=ios) lasso(1:3,(ires-1)*5+4)
      if (ios/=0) stop "BUG parsing ATOM line"
    case (" CB ")
      read(aline(31:54),'(3f8.3)',iostat=ios) lasso(1:3,(ires-1)*5+5)
      if (ios/=0) stop "BUG parsing ATOM line"
    end select
  enddo
  !! diagnostic
  ! write(0,*) lasso(:,:)
  close(11)
  write(*,'(i9," Residues read from ",a)') nres, trim(lassofile)
  !! ------------------------------------- read ligand
  !! Assign Hbonding types 0=neither,1=acceptor,2=donor,3=both
  open(lunit,file=trim(ligandfile),status='old',form='formatted',iostat=ios)
  if (ios/=0) stop 'anchor.f90:: ERROR ligandfile not found'
  !! read ligand coordinates
  iat = 0
  j = 0
  do 
    read(lunit,'(a)',iostat=ios) aline
    if (ios/=0) exit
    if (aline(1:5)/="ATOM ") cycle
    call names_gethtype(resname=aline(18:21),atomname=aline(13:16),htype=ht)
    if (ht < 0 ) then
      iu = iu + 1
      cycle
    endif
    read(aline(31:54),'(3f8.3)',iostat=ios) atvec
    if (ios/=0) stop 'anchor.f90:: error parsing ATOM line. line 112'
    dd = 999.
    do i=1,5*nres
      vec  = atvec - lasso(1:3,i)
      dist = sqrt(sum(vec*vec))
      if (dist < dd) dd = dist
    enddo
    if (dd > MAXD) then
      j = j + 1
      cycle
    endif
    iat = iat + 1
  enddo
  nat = iat
  rewind(lunit)
  write(*,'("Reading ",i9," atoms from ",a)') nat, trim(ligandfile)
  write(*,'(i9," atoms outside bounds ")') j
  write(*,'("WARNING: ",i9," unrecognized atoms ignored ")') iu
  allocate(ligand_coords(3,nat),ligand_htype(nat),stat=ios)
  if (ios/=0) stop 'anchor:: error allocating ligand'
  iat = 0
  do 
    read(lunit,'(a)',iostat=ios) aline
    if (ios/=0) exit
    if (aline(1:5)/="ATOM ") cycle
    call names_gethtype(resname=aline(18:21),atomname=aline(13:16),htype=ht)
    if (ht < 0 ) cycle
    read(aline(31:54),'(3f8.3)',iostat=ios) atvec
    if (ios/=0) stop 'anchor.f90:: error parsing ATOM line. line 112'
    dd = 999.
    do i=1,5*nres
      vec  = atvec - lasso(1:3,i)
      dist = sqrt(sum(vec*vec))
      if (dist < dd) dd = dist
    enddo
    if (dd > MAXD) cycle
    iat = iat + 1
    read(aline(31:54),'(3f8.3)',iostat=ios) ligand_coords(1:3,iat)
    ligand_htype(iat) = ht
  enddo
  !!----------------- if H-bond is already made in ligand, treat it as a new type
  do iat=1,nat-1
    if (ligand_htype(iat)==0) cycle
    do jat=iat+1,nat
      if (ligand_htype(jat)==0) cycle
      if (ligand_htype(jat)/=3 .and. ligand_htype(iat)==ligand_htype(jat)) cycle
      vec = ligand_coords(1:3,iat) - ligand_coords(1:3,jat)
      dd = sqrt(sum(vec*vec))
      if (dd > 3.6) cycle
      ht = 10*ligand_htype(iat) + ligand_htype(jat)
      select case (ht)
      case (12,13,21,23,31,32) !! donor-acceptor pair 
        ligand_htype(iat) = 4
        ligand_htype(jat) = 4
      end select
    enddo
  enddo
  write(*,'(i9," unsatisfied H-bond acceptors.")') count(ligand_htype==1)
  write(*,'(i9," unsatisfied H-bond donors.")') count(ligand_htype==2)
  write(*,'(i9," unsatisfied H-bond donor/acceptor atoms.")') count(ligand_htype==3)
  write(*,'(i9," satisfied H-bond donor/acceptor atoms.")') count(ligand_htype==4)
  !!----------------- initialize lassostack
  do i=1,STACKDEPTH+1
    allocate(lassostack(i)%coords(3,5*nres),stat=ios)
    if (ios/=0) stop 'anchor:: error allocating lassostack'
    lassostack(i)%score = -999
    lassostack(i)%irec = 0
  enddo
  lassostack(1)%coords = lasso
  call lassoscore(segment=lasso,NS=nres,atoms=ligand_coords,NA=nat, &
                  hbtype=ligand_htype,score=hbscore)
  lassostack(1)%score = 0.0
  lassostack(1)%irec = 0
  !!----------------- cycle through the database of backbone coords
  imdl = 0
  lastbest = -999.
  PASSLOOP: do
    nrmsd = 0
    imdl = imdl + 1
    write(0,*) "Starting model ",imdl
    irec = 1
    do
      nseg = nres
      call bdc_getnextsegment(irec=irec,segment=segment,slen=nseg,srec=srec)
      ! write(0,'("------------------------------",i12)') irec
      if (irec < 2) exit
      !!--- slide window across segment
      do i=1,nseg-nres+1
        ! segwindow(1:3,1:nres*5) = segment(1:3,(i-1)*5+1:(i+nres-1)*5)
        segwindow(:,:) = segment(:,(i-1)*5+1:(i+nres-1)*5)
        do k=1,nres
          cawindow(1:3,k) = segwindow(1:3,(k-1)*5+2)
        enddo
        hbest = -100
        do j=1,STACKDEPTH
          if (lassostack(j)%score < -100) cycle
          do k=1,nres
            calasso(1:3,k) = lassostack(j)%coords(1:3,(k-1)*5+2)
          enddo
          !! note: both CAs (cawindow) and all backbone (segwindow) are
          !! rotated by the same matrix, vector (mtx, vx).
          call superimp(calasso,cawindow,rmsd,natm=nres,mtx=mtx,vx=vx)
          call moveall(atoms=segwindow,nat=5*nres,mtx=mtx,vec=vx)
          !! diagnostic
          !write(0,'(i9,f8.3)') srec+i-1,rmsd
          if (rmsd > rmsdcut) cycle
          if (rmsd < rmsdmin) cycle
          nrmsd = nrmsd + 1
          segcenter = segwindow
          bestscore = -100.
          do ii=1,7
            select case (ii)
             case(1)
               wigglevec = (/0.,0.,0./)
             case(2)
               wigglevec = (/-wiggleroom,0.,0./)
             case(3)
               wigglevec = (/wiggleroom,0.,0./)
             case(4)
               wigglevec = (/0.,-wiggleroom,0./)
             case(5)
               wigglevec = (/0.,wiggleroom,0./)
             case(6)
               wigglevec = (/0.,0.,-wiggleroom/)
             case(7)
               wigglevec = (/0.,0.,wiggleroom/)
            end select
            do jj=1,nres*5
               if (segcenter(1,jj)<998.) segwindow(1:3,jj) = segcenter(1:3,jj) + wigglevec
            enddo
            call lassoscore(segment=segwindow,NS=nres,atoms=ligand_coords,NA=nat, &
                          hbtype=ligand_htype,score=hbscore)
            if (hbscore > bestscore) then
              bestscore = hbscore
              bestwiggle = wigglevec
            endif
          enddo
          hbscore = bestscore
          if (hbscore < lassostack(STACKDEPTH)%score) cycle
          if (hbscore > hbest) then
            do jj=1,nres*5
               if (segcenter(1,jj)<998.) lasso(1:3,jj) = segcenter(1:3,jj) + bestwiggle
            enddo
            lasso = segwindow
            hbest = hbscore
          endif
        enddo
        if (hbest < -99) cycle
        !! save best lasso
        k = STACKDEPTH+1
        do while (lassostack(k-1)%score < hbest)
          lassostack(k)%coords = lassostack(k-1)%coords
          lassostack(k)%score = lassostack(k-1)%score
          lassostack(k)%irec = lassostack(k-1)%irec
          k = k - 1
          if (k==1) exit
        enddo
        lassostack(k)%coords = lasso
        lassostack(k)%score = hbest
        lassostack(k)%irec = srec + i - 1
        !! diagnostic
        write(0,'(99f8.3)') lassostack(1:)%score
      enddo
    enddo
    !! ---------- write coordinates to movie file
    !! first copy ligand
    rewind(lunit)
    write(munit,'(a,i9)') "MODEL ",imdl
    write(munit,'(a,i9)') "REMARK Ligand atoms"
    do 
      read(lunit,'(a)',iostat=ios) aline
      if (ios/=0) exit
      if (aline(1:5)/="ATOM ") cycle
      aline(22:22) = "X"
      write(munit,'(a)') trim(aline)
    enddo
    write(munit,'(a)') "TER "
    aline = "ATOM      1  N   ALA Z   1      -9.841  49.939  14.888  1.00  0.00"
    write(*,'(i9," peptides had RMSD<",f8.3," and >",f8.3)') nrmsd,rmsdcut,rmsdmin
    do k=1,STACKDEPTH
      aline(22:22) = achar(ichar("A") + k - 1)
      write(*,'("Lassoenergy for MODEL ",i3," number ",i3," = ",f9.2)') imdl,k,lassostack(k)%score
      write(munit,'(a,f9.2)') "REMARK quality score = ",lassostack(k)%score
      do i=1,nres
        !! N
        write(aline(23:26),'(i4)') i
        write(aline(7:11),'(i5)') k*1000+5*(i-1)+1
        write(aline(31:54),'(3f8.3)') lassostack(k)%coords(1:3,(i-1)*5+1)
        aline(13:16) = " N  "
        write(munit,'(a)') trim(aline)
        !! CA
        write(aline(7:11),'(i5)') k*1000+5*(i-1)+2
        write(aline(31:54),'(3f8.3)') lassostack(k)%coords(1:3,(i-1)*5+2)
        aline(13:16) = " CA "
        write(munit,'(a)') trim(aline)
        !! C
        write(aline(7:11),'(i5)') k*1000+5*(i-1)+3
        write(aline(31:54),'(3f8.3)') lassostack(k)%coords(1:3,(i-1)*5+3)
        aline(13:16) = " C  "
        write(munit,'(a)') trim(aline)
        !! O
        write(aline(7:11),'(i5)') k*1000+5*(i-1)+4
        write(aline(31:54),'(3f8.3)') lassostack(k)%coords(1:3,(i-1)*5+4)
        aline(13:16) = " O  "
        write(munit,'(a)') trim(aline)
        if (lassostack(k)%coords(1,(i-1)*5+5) < 998.) then
          !! CB
          write(aline(7:11),'(i5)') k*1000+5*(i-1)+5
          write(aline(31:54),'(3f8.3)') lassostack(k)%coords(1:3,(i-1)*5+5)
          aline(13:16) = " CB "
          write(munit,'(a)') trim(aline)
        endif
      enddo
      write(*,'("Lassoenergy for MODEL ",i3," number ",i3," = ",f9.2)') imdl,k,lassostack(k)%score
      write(munit,'(a)') "TER "
    enddo
    write(munit,'(a,i9)') "ENDMDL",imdl
    !! new pass must improve score by at least a little (0.2)
    if (lassostack(1)%score < lastbest+0.2) exit
    lastbest = lassostack(1)%score
  enddo PASSLOOP
  !! ---------------- clean up 
  close(lunit)
  close(munit)
  call bdc_close(segment)
  deallocate(segwindow,segcenter,lasso,calasso,cawindow)
  do i=1,STACKDEPTH+1
    deallocate(lassostack(i)%coords)
  enddo
  deallocate(ligand_coords,ligand_htype)
CONTAINS
  subroutine lassoscore(segment,NS,atoms,NA,hbtype,score)
    implicit none
    integer,intent(in) :: NS, NA
    real,dimension(3,NS*5),intent(in) :: segment
    real,dimension(3,NA),intent(in) :: atoms
    integer,dimension(NA),intent(in) :: hbtype
    real,intent(out) :: score
    real,parameter :: hbmade=2.0,hbblocked=-2.0,npcontact=0.2
    real,parameter :: hbcut=3.5,npcut=5.8,blockcut=4.0,tooclose=3.0
    integer :: i,j,k,hb
    real :: dd
    real,dimension(3) :: vec
    !! ------
    ! write(0,*) segment
    ! stop 'in lassoscore'
    score = 0.
    do k=1,NA 
      hb = hbtype(k)
      do i=1,NS 
        do j=1,5
          if (segment(1,(i-1)*5+j)>998.) cycle
          vec = segment(1:3,(i-1)*5+j) - atoms(1:3,k)
          dd = sum(vec*vec)
            !! diagnostic
          ! write(0,'("dd=  ",f12.3,3i8)') dd,k,i,j
          if (dd > npcut*npcut) cycle
          dd = sqrt(dd)
          if (dd<tooclose) then
            !! diagnostic
            !write(0,'("too close ",f12.3,3i8)') dd,k,i,j
            score = -900
            return
          endif
          !! 10*hb = 0:nonpolar,10:acceptor,20:donor,30:both,40:satisfied
          !! j     = 1:donor,2:nonpolar,3:nonpolar,4:acceptor,5:nonpolar (N,CA,C,O,CB)
          select case (hb*10+j)
          case (1,4,12,13,14,15,21,22,23,25,32,33,35)  !! nonpolar + donor or acceptor
            if (dd<blockcut) score = score + hbblocked
          case (2,3,5,41,42,43,44,45)  !! nonpolar + nonpolar, nonpolar + satisfied
            if (dd<npcut) score = score + npcontact
          case (11,24,31,34)  !! acceptor + donor
            if (dd<hbcut) score = score + hbmade
          case default
            write(0,*) "BUG WARNING! hb=",hb," j=",j," case=",hb*10+j
          end select
        enddo
      enddo
    enddo
    ! write(0,'("lassoscore result ",f12.3)') score
    return
  end subroutine lassoscore
  
end program  anchor
