module bdc
  implicit none
  private
  integer,parameter :: RECSIZE=5*4*3+4
  integer,save :: bunit=12
  integer :: irec, ios=0
  character(len=200) :: binfile
  !!----------
  public :: bdc_getnextsegment, bdc_init, bdc_close
CONTAINS
    !! ---------------------------------------------------------
    subroutine bdc_init(filename)
      character(len=*) :: filename
      binfile = trim(filename)
    end subroutine bdc_init
    !! ---------------------------------------------------------
    subroutine bdc_close(segment)
      real,dimension(:,:),pointer,optional :: segment
      logical :: yet=.false.
      inquire(unit=bunit,opened=yet)
      if (yet) close(bunit)
      if (present(segment)) then
        if (associated(segment)) deallocate(segment)
      endif
    end subroutine bdc_close
    !! ---------------------------------------------------------
    subroutine bdc_getnextsegment(irec,segment,slen,srec)
      implicit none
      integer,intent(inout) :: irec
      real,dimension(:,:),pointer :: segment
      integer,intent(inout) :: slen
      integer,optional,intent(out) :: srec 
      logical :: yet=.false.
      real,dimension(3,5) :: fiveatom
      integer :: istart,iend,nres,resid,i,j,k
      !!---
      nres = slen
      if (associated(segment)) deallocate(segment)
      inquire(unit=bunit,opened=yet)
      if (.not.yet) then
        open(bunit,file=trim(binfile),status='old',form='unformatted',access='direct',recl=RECSIZE,iostat=ios)
        if (ios/=0) stop 'Error opening binary file'
      endif
      do
        read(bunit,rec=irec,iostat=ios) resid,fiveatom
        resid = 0
        do while (resid <= 1000) 
          irec = irec + 1
          read(bunit,rec=irec,iostat=ios) resid,fiveatom
          if (ios/=0) exit
        enddo
        istart = irec
        do while (resid > 1000) 
          irec = irec + 1
          read(bunit,rec=irec,iostat=ios) resid,fiveatom
          if (ios/=0) exit
        enddo
        iend = irec - 1
        slen = iend - istart + 1
        if (ios/=0) then
           irec = 1   !! rewinding
           if (slen < nres) return
        endif
        if (slen >= nres) exit
      enddo
      if (present(srec)) srec = istart
      allocate(segment(3,5*slen),stat=ios) ; if (ios/=0) stop 'bdc:: Error allocating segment'
      do i=1,slen
        j = istart + i - 1
        read(bunit,rec=j,iostat=ios) resid,segment(1:3,(i-1)*5+1:(i*5))
        if (ios/=0) stop 'bdc:: Error reading bdc file'
      enddo
    end subroutine bdc_getnextsegment
        
end module bdc

